#region Using declarations

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Linq;
using System.Xml.Serialization;
using NinjaTrader.Cbi;
using NinjaTrader.Data;
using NinjaTrader.Gui.Chart;

#endregion Using declarations

namespace NinjaTrader.Indicator {

	[Description("Vitalij's Level II Indicator for NinjaTrader 7." + "\r\n" + "https://bitbucket.org/ninjatrader")]
	public class VitLevel2 : Indicator {

		#region Variables

		Dictionary<double, long> rowsAsk = new Dictionary<double, long>();
		Dictionary<double, long> rowsBid = new Dictionary<double, long>();

		int maxBarWidth = 100;
		int maxBarHeight = 50;
		int minBarHeight = 3;

		string error = "No market depth data available!";
		Font textFont = new Font("Arial", 8, FontStyle.Regular);

		StringFormat sfRightCenter = new StringFormat();
		StringFormat sfLeftCenter = new StringFormat();
		StringFormat sfLeftBottom = new StringFormat();
		StringFormat sfLeftTop = new StringFormat();

		int opacityAskBars = 50;
		int opacityBidBars = 50;

		Color colorAskBars = Color.Plum;
		Color colorBidBars = Color.LightGreen;
		Color colorAskText = Color.Black;
		Color colorBidText = Color.Black;
		Color colorAskTextSum = Color.Black;
		Color colorBidTextSum = Color.Black;

		Brush brushAskBar;
		Brush brushBidBar;
		Brush brushAskText;
		Brush brushBidText;
		Brush brushAskTextSum;
		Brush brushBidTextSum;

		#endregion Variables

		#region Properties

		[Category("Size")]
		[NinjaTrader.Gui.Design.DisplayName("Bar width (max.)")]
		public int MaxBarWidth {
			get { return maxBarWidth; }
			set { maxBarWidth = Math.Max(Math.Min(value, int.MaxValue), 1); }
		}

		[Category("Size")]
		[NinjaTrader.Gui.Design.DisplayName("Bar height (max.)")]
		public int MaxBarHeight {
			get { return maxBarHeight; }
			set { maxBarHeight = Math.Max(Math.Min(value, int.MaxValue), minBarHeight); }
		}

		[Category("Size")]
		[NinjaTrader.Gui.Design.DisplayName("Bar height (min.)")]
		public int MinBarHeight {
			get { return minBarHeight; }
			set { minBarHeight = Math.Max(Math.Min(value, maxBarHeight), 1); }
		}

		[XmlIgnore()]
		[Description("Text font")]
		[Category("Font")]
		[NinjaTrader.Gui.Design.DisplayName("Text font")]
		public Font TextFont {
			get { return textFont; }
			set { textFont = value; }
		}

		[Browsable(false)]
		public string TextFontSerialize {
			get { return NinjaTrader.Gui.Design.SerializableFont.ToString(textFont); }
			set { textFont = NinjaTrader.Gui.Design.SerializableFont.FromString(value); }
		}

		[XmlIgnore()]
		[Category("Ask colors")]
		[NinjaTrader.Gui.Design.DisplayName("Bar color")]
		public Color ColorAskBars {
			get { return colorAskBars; }
			set { colorAskBars = value; }
		}

		[Browsable(false)]
		public string ColorAskBarsSerialize {
			get { return NinjaTrader.Gui.Design.SerializableColor.ToString(colorAskBars); }
			set { colorAskBars = NinjaTrader.Gui.Design.SerializableColor.FromString(value); }
		}

		[XmlIgnore()]
		[Category("Ask colors")]
		[NinjaTrader.Gui.Design.DisplayName("Text color (Volume)")]
		public Color ColorAskText {
			get { return colorAskText; }
			set { colorAskText = value; }
		}

		[Browsable(false)]
		public string ColorAskTextSerialize {
			get { return NinjaTrader.Gui.Design.SerializableColor.ToString(colorAskText); }
			set { colorAskText = NinjaTrader.Gui.Design.SerializableColor.FromString(value); }
		}

		[XmlIgnore()]
		[Category("Ask colors")]
		[NinjaTrader.Gui.Design.DisplayName("Text color (Sum)")]
		public Color ColorAskTextSum {
			get { return colorAskTextSum; }
			set { colorAskTextSum = value; }
		}

		[Browsable(false)]
		public string ColorAskTextSumSerialize {
			get { return NinjaTrader.Gui.Design.SerializableColor.ToString(colorAskTextSum); }
			set { colorAskTextSum = NinjaTrader.Gui.Design.SerializableColor.FromString(value); }
		}

		[Category("Ask colors")]
		[NinjaTrader.Gui.Design.DisplayName("Bar opacity (%)")]
		public int OpacityAskBars {
			get { return opacityAskBars; }
			set { opacityAskBars = Math.Max(Math.Min(value, 100), 0); }
		}

		[XmlIgnore()]
		[Category("Bid colors")]
		[NinjaTrader.Gui.Design.DisplayName("Bar color")]
		public Color ColorBidBars {
			get { return colorBidBars; }
			set { colorBidBars = value; }
		}

		[Browsable(false)]
		public string ColorBidBarsSerialize {
			get { return NinjaTrader.Gui.Design.SerializableColor.ToString(colorBidBars); }
			set { colorBidBars = NinjaTrader.Gui.Design.SerializableColor.FromString(value); }
		}

		[XmlIgnore()]
		[Category("Bid colors")]
		[NinjaTrader.Gui.Design.DisplayName("Text color (Volume)")]
		public Color ColorBidText {
			get { return colorBidText; }
			set { colorBidText = value; }
		}

		[Browsable(false)]
		public string ColorBidTextSerialize {
			get { return NinjaTrader.Gui.Design.SerializableColor.ToString(colorBidText); }
			set { colorBidText = NinjaTrader.Gui.Design.SerializableColor.FromString(value); }
		}

		[XmlIgnore()]
		[Category("Bid colors")]
		[NinjaTrader.Gui.Design.DisplayName("Text color (Sum)")]
		public Color ColorBidTextSum {
			get { return colorBidTextSum; }
			set { colorBidTextSum = value; }
		}

		[Browsable(false)]
		public string ColorBidTextSumSerialize {
			get { return NinjaTrader.Gui.Design.SerializableColor.ToString(colorBidTextSum); }
			set { colorBidTextSum = NinjaTrader.Gui.Design.SerializableColor.FromString(value); }
		}

		[Category("Bid colors")]
		[NinjaTrader.Gui.Design.DisplayName("Bar opacity (%)")]
		public int OpacityBidBars {
			get { return opacityBidBars; }
			set { opacityBidBars = Math.Max(Math.Min(value, 100), 0); }
		}

		#endregion Properties

		#region Methods

		protected override void Initialize() {
			CalculateOnBarClose = true;
			Overlay = true;
			AutoScale = false;

			Name = "Level II Indicator";

			sfLeftCenter.Alignment = StringAlignment.Far;
			sfLeftCenter.LineAlignment = StringAlignment.Center;
			sfRightCenter.Alignment = StringAlignment.Near;
			sfRightCenter.LineAlignment = StringAlignment.Center;
			sfLeftBottom.Alignment = StringAlignment.Far;
			sfLeftBottom.LineAlignment = StringAlignment.Near;
			sfLeftTop.Alignment = StringAlignment.Far;
			sfLeftTop.LineAlignment = StringAlignment.Far;

			brushAskText = new SolidBrush(colorAskText);
			brushBidText = new SolidBrush(colorBidText);
			brushAskTextSum = new SolidBrush(colorAskTextSum);
			brushBidTextSum = new SolidBrush(colorBidTextSum);
			brushAskBar = new SolidBrush(Color.FromArgb(opacityAskBars * 255 / 100, colorAskBars));
			brushBidBar = new SolidBrush(Color.FromArgb(opacityAskBars * 255 / 100, colorBidBars));
		}

		protected override void OnMarketDepth(MarketDepthEventArgs ev) {
			if (ev.MarketDataType == MarketDataType.Ask) {
				SetMarketDepthVolume(ev, rowsAsk);
			} else if (ev.MarketDataType == MarketDataType.Bid) {
				SetMarketDepthVolume(ev, rowsBid);
			}
		}

		protected void SetMarketDepthVolume(MarketDepthEventArgs ev, Dictionary<double, long> rows) {
			if (ev.Price == 0) {
				rowsAsk.Clear();
				rowsBid.Clear();
			}

			if (ev.Operation == Operation.Remove) {
				rows.Remove(ev.Price);
			} else /* if (ev.Operation == Operation.Insert || ev.Operation == Operation.Update) */ {
				rows[ev.Price] = ev.Volume;
			}

			ChartControl.ChartPanel.Invalidate();
		}

		public override void Plot(Graphics graphics, Rectangle bounds, double min, double max) {
			// if right-most bar is not a real-time updating bar, then do not display Level II histogram
			if (ChartControl.LastBarPainted < Bars.Count - 1) {
				return;
			}

			// display error message
			if (!DisplayTime() || rowsAsk.Count < 1 || rowsBid.Count < 1) {
				graphics.DrawString(error, ChartControl.Font, new SolidBrush(ChartControl.AxisColor), bounds.Right, bounds.Bottom, sfLeftTop);
				return;
			}

			// sum up volume
			long maxAskVolume = rowsAsk.Max(v => v.Value);
			long maxBidVolume = rowsBid.Max(v => v.Value);

			// calc scale/dimensions/size
			float barWidthScale = (float)maxBarWidth / Math.Max(maxAskVolume, maxBidVolume);
			int[] y = GetYsByPrice(TickSize);
			bool isCramped = (textFont.GetHeight(graphics) > y[4]);

			// draw ask bars
			foreach (KeyValuePair<double, long> row in rowsAsk) {
				int barWidth = (int)Math.Round(row.Value * barWidthScale);
				y = GetYsByPrice(row.Key);

				graphics.FillRectangle(brushAskBar, bounds.Right - barWidth, y[0], barWidth, y[2]);
				if (!isCramped) {
					graphics.DrawString(row.Value.ToString(), textFont, brushAskText, bounds.Right, y[3], sfLeftCenter);
				} else if (row.Value == maxAskVolume) {
					graphics.DrawString(row.Value.ToString(), textFont, brushAskText, bounds.Right - barWidth, y[3], sfRightCenter);
				}
			}

			// draw bid bars
			foreach (KeyValuePair<double, long> row in rowsBid) {
				int barWidth = (int)Math.Round(row.Value * barWidthScale);
				y = GetYsByPrice(row.Key);

				graphics.FillRectangle(brushBidBar, bounds.Right - barWidth, y[0], barWidth, y[2]);
				if (!isCramped) {
					graphics.DrawString(row.Value.ToString(), textFont, brushBidText, bounds.Right, y[3], sfLeftCenter);
				} else if (row.Value == maxBidVolume) {
					graphics.DrawString(row.Value.ToString(), textFont, brushBidText, bounds.Right - barWidth, y[3], sfRightCenter);
				}
			}

			// draw ask sum
			y = GetYsByPrice(rowsAsk.Keys.Max() + TickSize);
			graphics.DrawString(rowsAsk.Sum(v => v.Value).ToString(), textFont, brushAskTextSum, bounds.Right, y[3], sfLeftCenter);

			// draw bid sum
			y = GetYsByPrice(rowsBid.Keys.Min() - TickSize);
			graphics.DrawString(rowsBid.Sum(v => v.Value).ToString(), textFont, brushBidTextSum, bounds.Right, y[3], sfLeftCenter);
		}

		private int[] GetYsByPrice(double price) {
			int barSpacing = 1; // TODO: should be configurable?

			int y0 = ChartControl.GetYByValue(this, price + TickSize);
			int y1 = ChartControl.GetYByValue(this, price);
			int height = y1 - y0 - barSpacing;
			int realHeight = Math.Max(minBarHeight, Math.Min(maxBarHeight, height));

			y0 += (height - realHeight);

			return new int[] { y0, y1, realHeight, y0 + (y1 - y0) / 2, height };
		}

		protected override void OnBarUpdate() {
		}

		protected override void OnTermination() {
			sfRightCenter.Dispose();
			sfLeftCenter.Dispose();
			sfLeftBottom.Dispose();
			sfLeftTop.Dispose();

			brushAskBar.Dispose();
			brushBidBar.Dispose();
			brushAskText.Dispose();
			brushBidText.Dispose();
			brushAskTextSum.Dispose();
			brushBidTextSum.Dispose();
		}

		#endregion Methods

		#region Miscellaneous - see @BarTimer.cs

		private DateTime Now {
			get {
				DateTime now = (Bars.MarketData.Connection.Options.Provider == Cbi.Provider.Replay ? Bars.MarketData.Connection.Now : DateTime.Now);

				if (now.Millisecond > 0)
					now = Cbi.Globals.MinDate.AddSeconds((long)System.Math.Floor(now.Subtract(Cbi.Globals.MinDate).TotalSeconds));

				return now;
			}
		}

		private bool DisplayTime() {
			if (ChartControl != null
					&& Bars != null
					&& Bars.Count > 0
					&& Bars.MarketData != null
					&& Bars.MarketData.Connection.PriceStatus == Cbi.ConnectionStatus.Connected
					&& Bars.Session.InSession(Now, Bars.Period, true, Bars.BarsType))
				return true;

			return false;
		}

		#endregion Miscellaneous - see @BarTimer.cs

	}
}
